/*
 * =================================================================================================
 *                                  VARIOUS GENERAL ROUTINES
 * =================================================================================================
 */
#include <ctime>
#include <climits>
#include <iostream>
#include "general.h"

//I understand "long long" means __int64 by default in window: I am not sure this is necessary
//for using the non-standard type "long long"
//#ifdef WIN32
//  #define unsigned_long_long unsigned __int64
//  #define long_long __int64
//#else // gcc. Might not work on other compilers!
//  #define unsigned_long_long unsigned long long
//  #define long_long long long
//#endif

//Attention: "long long" is not a standard type (but c99), see also above for Windows versions
long long        tStart; 
long long        cumulatedTimePassed;

void startTimer(){
    tStart=(long long) clock();
    cumulatedTimePassed = 0;
}

//Returns the time passed since tStart: attention to the "wrap around" issued described in "man 3
//clock": on a 32 bits system, the value returned by clock is the same every ~70 minutes=2INT_MAX
//ticks. As such, on certain systems, this function only measures times inferior to ~70 minutes!!
long long timeCpuPassedMSec(){
    long long tEnd =  clock();          //attention ``long long'' is not a standard
    long long tStartCpy, tEndCpy;
    tStartCpy = tStart;                 //Keep here the initial tStart value for future
    tEndCpy   = tEnd;
    //ADDRESSING issue ''wrap around' described in ``man 3 clock'': "Note  that  the  time can wrap around."
    //This is not the case on all systems, but on my 32bit system, the counter goes as follows:
    //0, 1, 2....INT_MAX-1, INT_MAX, -INT_MAX, -INT_MAX -1, ... , 0, 1, 2 ...
    //If INT_MAX=2147483648, CLOCKS_PER_SEC=10^6(POSIX requirement), the first sign change arises after 2147s.
    if(tEnd<0&&tStart>0){               //After INT_MAX ticks, clock_t() returns "int" values<0
        tEnd   = INT_MAX + tEnd  ;      //ticks after passing INT_MAX threshold
        tStart = tStart - INT_MAX;      //- (ticks from start to INT_MAX)
    }                                   
    else if(tEnd>0&&tStart>0&&tEnd<tStart){//Deal with case: clock() goes over 2*INT_MAX
        tEnd = tEnd + INT_MAX;
    } 
    long long timePassed = (tEnd - tStart) / ((long long)(CLOCKS_PER_SEC/1000));
    
    int updateCumTm = 0;                //if 1, we signal the need for updating cumulatedTimePassed
    tStart = tStartCpy;                 //tStart was modified above only to find current timePassed
    if(timePassed<0){
        std::cerr<<"Warning: wrap-around int in clock(): replace "<<timePassed<<" with 0"
            <<"tStart="<<tStartCpy
            <<"tEnd="<<tEndCpy<<std::endl;
        updateCumTm = 1;
    }
    //Add a cumulated time passed, and allow the real timer to re-start via tStart update below
    if(timePassed>10000)
        updateCumTm = 1;
    timePassed += cumulatedTimePassed;
    if(updateCumTm){
        tStart=(long long) clock();
        cumulatedTimePassed = timePassed;
    }
	return timePassed;
}

//int to string, a C++ classical
std::string toString(int number){
    std::ostringstream convert;   
    convert << number;      
    return convert.str(); 
}
