#include "config.h"

#include <iostream>
#include <iomanip>
#include <cassert>
#include <climits>
#include <cstdlib>
#include <cmath>
#include <string>
using namespace std;

#include "globals.h"
#include "colGenerator.h"
#include "irm.h"
#include "general.h" 

#define NORAY_FOUND -1
#define ceil std::ceil
#define floor std::floor

/*************************************************************************************************/
//The only problem-specific routines to be found when linking, using namespace problemSpecific
#include "subprob.h"
/*************************************************************************************************/

colGenerator*cg;                                 
double lbVal;
double ubVal;
double t;                                        //last returned t
double tNew;                                     //tNew in the paper
double ca;
int    p ;
double caNew;
int    pNew;
double*ub;                                       //upper bound solution
double*rDelt;                                    //ub-lb = ub - tr

int jCoord;                                      //coordinate axis/index that maximizes below
int maxIntsBtwn;                                 //number of integer points on that axis betwn lb,ub

int * r;
int * rNew;                                      //r' in the paper
int * a;

int failed =  0;                                 //Number of runs failing within allowed time
int successes = 0;                               //Nr. of success runs within allowed time

template <class T1, class T2>
double scalProd(T1* v1, T2* v2){
   double prod = 0;
   for(int i=0;i<n;i++)
       prod+=(v1[i]*v2[i]); 
   return prod;
}

void initGlobDataStr(){
   r     = new int[n]; 
   rNew  = new int[n]; 
   a     = new int[n]; 
   ub    = new double[n];
   rDelt = new double[n];
   lbVal = -INT_MAX;
   ubVal = INT_MAX;
   cg = new colGenerator(n);
   cg->setObjective(b);
   cg->setVarBounds(l,u);
}
void delGlobDataStr(){
   if(a!=NULL)
        delete[] a;
   delete[] r;
   delete[] rNew;
   delete[] ub;
   delete[] rDelt;
   delete cg;
}
double rayFromActUBRefined = 1;
//Returns the ray by rounding the actual upper bound with no t-scaling
//Input: global ub, Returned: 1 if new ray exists or 0 otherwise; returned via globals: rNew
int rayFromActUB(){
    int rSum = 0;
    clog<<"rnew0: ";                         
    for(int i=0;i<n;i++){           
        rNew[i] = floor(rayFromActUBRefined*ub[i]-RC_EPS);  //Input: r, t, rDelt, ub
        clog<<rNew[i]<<"("<<ub[i]<<")"<<", ";//Output: rNew
        rSum += rNew[i];
    }
    if(scalProd(b,rNew)<=0)                //rare exception: due to integer rounding, new ray
        return 0;                         //can be non-progressive w. r. t. the objective function
    return 1;
}

//Returns a ray by rounding t*ub to the closest integer on each position
//Returns the ray by rounding the actual upper bound with t-scaling, or 0 if no ray found this way
//Input: global ub, Returned: 2; returned via globals: rNew
int rayFromScaledUB2(int p, double ca){
    if(lbVal<RC_EPS){
        clog<<"No feasible ray found yet";
        return rayFromActUB();
    }
    double correction=1;
    if(ubVal/lbVal>3)
        correction = ubVal/(2*lbVal);
    clog<<"Ray constructed from UB: p="<<p<<", ca="<<ca<<endl;
    for(int i=0;i<n;i++){                    
        rNew[i] = (ub[i]*p/ca < r[i])? ceil(ub[i]*p/ca) : floor(ub[i]*p/ca);
        rNew[i] = rNew[i]/correction;
        clog<<rNew[i]<<"("<<i<<":"<<ub[i]*p/ca<<",r="<<r[i]<<")"<<", ";
    }
    clog<<endl;
    if(scalProd(b,rNew)<=0)                //rare exception: due to integer rounding, new ray
        return 0;                         //can be non-progressive w. r. t. the objective function
    return 2;
}

//Returns a ray by rounding (t*ub+r)/2 to the closest integer on each position
//Returns the ray by rounding the actual upper bound with no t-scaling
//Input: global ub, Returned: 1 if new ray exists or 0 otherwise; returned via globals: rNew
int rayFromMidLbUb(){
    return 0;                                   //DISABLE
    int pp= 1;                                  //put =c/2+p/2 if you want interior points 
    clog<<"FIN rnew2: ";
    for(int i=0;i<n;i++){                    
        rNew[i] = floor((r[i]+ub[i]*pp/ca)/2+0.5);
        clog<<rNew[i]<<"("<<r[i]<<"-"<<ub[i]*pp/ca<<")"<<", ";
    }
    clog<<endl;
    return 1;
}

//Input : p and ca that yield the t here used
//Return: idxOfMaxDelta - the coordinate where the number of integers is maximum
//        maxIntsBtwn   - the number of integer values between r (exclusive) and the closest 
//                        integer to t Ub on returned coordinate idxOfMaxDelta
void maxIntsBetweenUb(int p, double ca, int& idxOfMaxDelta, int& maxIntsBtwn){
    maxIntsBtwn = 0;                               //max integer coordinates between ub and lb on any index
    for(int i=0;i<n;i++)
        if(abs(floor(ub[i]*p/ca+0.5)-r[i])>maxIntsBtwn){
            idxOfMaxDelta = i;
            maxIntsBtwn = abs(floor(ub[i]*p/ca+0.5)-r[i]);
        } 
   clog<<"FILTER"<<"maxIntsBtwn="<<maxIntsBtwn<<endl; 
}
//Input: orderNo - order number in the queue of rays to be tried, 0-indexed
//               - the last two parameters (yielding t) can take their value from the global ones
//Return value: the order number of the ray that could be generated, ie.,
//              it can be the same as the ray at order number orderNo is feasible
// 
int newIntegerRay(int orderNo, int p, double ca){
    if(orderNo==0) orderNo++; 
    //if(orderNo==0){                             //at the first call, try first a direction close to
    //    if(rayFromActUB())                      //the actual upper bound: for initial constraints
    //        return orderNo;                     //constraints -> ADD it to the paper ADD ADD    
    //    orderNo++;
    //}
    if(orderNo==1){                               //second call
        if(rayFromScaledUB2(p, ca))
            return orderNo;
        orderNo++;
    }
    if(ca<RC_EPS){                              
        do{
            rayFromActUBRefined *=1.2;
            clog<<"rayFromActUBRefined="<<rayFromActUBRefined<<endl;
            rayFromActUB(); 
        }while(scalProd(b,rNew)<=0);
        return 0;
        //cerr<<"Not yet implemented ca=0 and the upper bound does not generate better rays\n";  
        //exit(EXIT_FAILURE); 
    }
    if(orderNo==2){                             //third call: midpoint(ub/t,lb/t)
      if(rayFromMidLbUb())
//        if(rayFromScaledUB2(p, ca))
            return orderNo;
        orderNo++;
    }
    //Considering intermediate points on r--->t ub
    if(orderNo>=3){
        if(orderNo==3){                        //calculate rDelt and max integer rDelt (interm points)
            for(int i=0;i<n;i++)
                rDelt[i] = ub[i]*p/ca - r[i];
            maxIntsBetweenUb(p,ca,jCoord,maxIntsBtwn);
            if(maxIntsBtwn==0){
                cerr<<"Warng: no integer coordinates between lb and ub -> refining\n";
                return NORAY_FOUND;
            }
        }
        double beta;                        //r'=r+beta*Delta 
        int jVal = orderNo-2;               //start at 1: abs value of y^\beta on coord jCoord
        if(jVal>=maxIntsBtwn)
            return NORAY_FOUND;
        if(rDelt[jCoord]>=0)
            beta = (jVal+0.5)/rDelt[jCoord];//at jCoord, r+beta*Delta=r+0.5+jVal if rDelt[j]>0
        else
            beta =-(jVal+0.5)/rDelt[jCoord];//at jCoord, r+beta*Delta=r+0.5+jVal if rDelt[j]>0
            
        clog<<"FIN rnew3: "<<jVal<<" steps with beta="<<beta<<endl;
        for(int i=0;i<n;i++){                    
            rNew[i] = floor(r[i]+beta*rDelt[i]);
            clog<<rNew[i]<<"("<<i<<"-"<<r[i]+beta*rDelt[i]<<"in["<<ub[i]*p/ca<<","<<r[i]<<"])"<<", ";
            if(rNew[i]<l[i]&&(l[i]==0)){
                    clog<<"Warning: tried negative r[i] value\n";
                    return NORAY_FOUND;       
            }
        }
        clog<<endl;
        if(scalProd(b,rNew)<=0)
            return NORAY_FOUND;
        return orderNo;
    }
    cerr<<"Warning: it should never reach this point (t is too small?)---orderNo="<<orderNo<<endl;
    return NORAY_FOUND;
}

//calls newIntegerRay using the current t=p/ca
int newIntegerRay(int orderNo){
    return newIntegerRay(orderNo, p, ca);
}

//output the number of IRM iterations and discretization refining calls
void printIterNumbers(string callsNewRay, string callsDisRef){
      if(callsNewRay.length()==0){
            //cout<<"     &        &       &";
            cout<<"\\multicolumn{3}{c||}{}&";
            return ;
      }
      cout<<setw(5)<<callsNewRay;
      cout<<"&"<<setw(7)<<callsDisRef<<" &";        
      cout<<setw(7)<<timeCpuPassedMSec()<<"&";
}

//output the current bound, usually the last characters of the output line
void printBounds(){
      cout<<"\\\\%base+lbVal,ubVal="<<setw(6)<<lBndBase<<"&"<<setw(6)
            <<setprecision(4)<<lbVal<<"," <<setw(7)<<ubVal <<"-> "
            <<setprecision(5)
            <<setw(5)<<ceil(lBndBase+lbVal)<<"&"<<setw(5)<<ceil(lBndBase+ubVal)<<"&"<<endl;
}
//Runs the algorithm, the data should have been loaded: n, b and l with loadData(..) in main.cpp
void irm(){
   startTimer(); 
   int callsNewRay = 0;
   int callsDisRef = 0;
   int gapPrinted  = 0;                                   //indicates if the 10% gap is already done
   int newLbFound, newUbFound;
   int rayNoInQueue;                                      //the order number of ray in queue to gen
   int rayNoGen;                                          //the order number really generated
   int i;
   initGlobDataStr();
   int rSum = 0; 
   
   for(int i=0;i<n;i++){
        r[i] = b[i];
        rSum += r[i];
   }
   if(rSum>1000*n){
        cout<<"This instance has a huge number of requests. Implement a scale-down when"
            <<"putting the start values b=r. Stop now"<<endl;
        exit(1);
   }
   if(rSum>10*n){
        clog<<"Scaling start r"<<endl;
        double scaleDwn = ((double)(10*n))/rSum;
        for(int i=0;i<n;i++)
            r[i] = b[i] * scaleDwn;
   }
        
    
   rayFromActUBRefined = 1;
   problemSpecific::irmSubproblem(r,t,a,ca,p);
   callsNewRay++;
   lbVal = scalProd(r,b) *ca/p;
   cg->modelAddColumn(a,ca);
   
   ubVal = cg->solve();
   cg->getDuals(ub);

   if( !(ceil(ubVal)>ceil(lbVal)&&(ubVal-lbVal>RC_EPS))){
      cerr<<"The instance seems very easy\n";
   } else
   do{
        //Try to generate a better ray
        rayNoInQueue = 1;
        do{
            clog<<"upper ="<<ubVal<<"(lbVal="<<lbVal<<") at"<<timeCpuPassedMSec()<<endl;
            rayNoGen = newIntegerRay(rayNoInQueue);       //queue from paper simulated using rayNoInQueue
            //HERE, YOU CAN TRY APPLYING newIntegerRay AGAIN WITH A DIFF p/ca PARAMS
            if(rayNoGen==NORAY_FOUND){                    //discretization refining lambda=2
                callsDisRef++;
                t=t/2;
                p=p*2;
                rSum = 0;
                for(i=0;i<n;i++){
                    r[i]*=2;
                    rSum += r[i];
                }
                rayNoInQueue = 1;
                newLbFound = 0;
                newUbFound = 0;
                clog<<"FILTERME Discretization Refining l="<<callsDisRef
                     <<" after solving a total of subprobs: "<<callsNewRay<<"\n";
            }else{
                rayNoInQueue = rayNoGen + 1;
                problemSpecific::irmSubproblem(rNew,tNew,a,caNew,pNew);//old a is lost here
                if(a!=NULL){
                    newLbFound = (   scalProd(rNew,b)*caNew/pNew > lbVal+RC_EPS);
                    newUbFound = (   scalProd(a,ub)     > caNew+RC_EPS);   //negative reduced cost
                }else{
                    newLbFound = 0;
                    newUbFound = 0;
                }
                callsNewRay++;
            }

        }while((!newLbFound)&&(!newUbFound)&&(timeCpuPassedMSec()<=MILISECS_MAX)) ;
        if(newUbFound){
            cg->modelAddColumn(a,caNew);
            ubVal = cg->solve();
            cg->getDuals(ub);
            clog<<"\nCFILTERnew upper found="<<ubVal<<"(lbVal="<<lbVal<<")"<<timeCpuPassedMSec()<<endl;
        }
        if(newLbFound){
            double scaleDwn=1;
            int rNewSum = 0;
            for(int i=0;i<n;i++)
                rNewSum+=rNew[i];
            if(rNewSum>2*rSum)          //quite extreme situation: can happen if lbVal very bad
                scaleDwn = 2;           //the objective is to avoid implicit discretiz. refining

            ca= caNew;
            p = pNew;
            t = tNew;
            for(int i=0;i<n;i++){      //Attention to r that is integer
                r[i] = rNew[i];
            }
            lbVal = scalProd(r,b)*ca/p ;
            clog<<"CFILTER new lower bound="<<lbVal<<","<<timeCpuPassedMSec()<<endl;
            ca= ca*scaleDwn;
            t = t *scaleDwn;
            for(int i=0;i<n;i++){      //Attention to r that is integer
                r[i] = r[i]/scaleDwn;
            }
        }
        if(!gapPrinted&&((lBndBase+ubVal)*9/10<=(lbVal+lBndBase))){
           gapPrinted = 1; 
           printIterNumbers(toString(callsNewRay), toString(callsDisRef));
        }     
        if(timeCpuPassedMSec()>MILISECS_MAX)
            break;
   }while( (ceil(ubVal)>ceil(lbVal-RC_EPS)) &&
           (ubVal-lbVal>RC_EPS)         );               //for cases like ubVal=5.0..01, lbVal=4.99..
   if(timeCpuPassedMSec()>MILISECS_MAX){  
        if(!gapPrinted)
            printIterNumbers("","");
        printIterNumbers("","");
        failed ++; 
   }else{
      if(!gapPrinted){
         printIterNumbers(toString(callsNewRay), toString(callsDisRef));
         gapPrinted = 1; 
      }     
      printIterNumbers(toString(callsNewRay), toString(callsDisRef));
      successes ++;
   }
   printBounds();
   //use below if you want to stop the program before finishing all instances in the file
   //if(failed>5 && failed > 3*successes){
   //      clog<<"Stoped after "<<failed<< "failures and only "<<successes<<"successes\n";
   //      exit(EXIT_FAILURE);
   //}
   clog<<"FINAL: lbVal="<<lbVal+RC_EPS<<" ubVal="<<ubVal<<endl;
   delGlobDataStr();
   clog<<"TIME="<<timeCpuPassedMSec()<<endl;
}
//launches IRM on al instances that can be read
void allInstIRM(char* inputFile){
    //instance name, globals: nr items, lower l upper u bounds, service required b
    while (problemSpecific::loadData(inputFile,n,l,u,b,lBndBase)){
        irm();
    }     
}
